import datetime as dt
from typing import Union, List


import pandas as pd

from mindsdb_sql.parser.dialects.mindsdb import CreateJob, DropJob
from mindsdb_sql.parser.ast import Identifier, Star, Select

from mindsdb_sdk.query import Query
from mindsdb_sdk.utils.sql import dict_to_binary_op
from mindsdb_sdk.utils.objects_collection import CollectionBase
from mindsdb_sdk.utils.context import set_saving


class Job:
    def __init__(self, project, name, data=None, create_callback=None):
        self.project = project
        self.name = name
        self.data = data

        self.query_str = None
        if data is not None:
            self._update(data)
        self._queries = []
        self._create_callback = create_callback

    def _update(self, data):
        # self.name = data['name']
        self.query_str = data['query']
        self.start_at = data['start_at']
        self.end_at = data['end_at']
        self.next_run_at = data['next_run_at']
        self.schedule_str = data['schedule_str']

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, query='{self.query_str}')"

    def __enter__(self):
        if self._create_callback is None:
            raise ValueError("The job is already created and can't be used to create context."
                               " To be able to use context: create job without 'query_str' parameter: "
                               "\n>>> with con.jobs.create('j1') as job:"
                               "\n>>>    job.add_query(...)")
        set_saving(f'job-{self.name}')

        return self

    def __exit__(self, type, value, traceback):
        set_saving(None)
        if type is None:
            if len(self._queries) == 0:
                raise ValueError('No queries were added to job')

            query_str = '; '.join(self._queries)

            self._create_callback(query_str)

            self.refresh()

    def refresh(self):
        """
        Retrieve job data from mindsdb server
        """
        job = self.project.get_job(self.name)
        self._update(job.data)

    def add_query(self, query: Union[Query, str]):
        """
        Add a query to job. Method is used in context of the job

        >>> with con.jobs.create('j1') as job:
        >>>    job.add_query(table1.insert(table2))

        :param query: string or Query object. Query.database should be emtpy or the same as job's project
        """
        if isinstance(query, Query):

            if query.database is not None and query.database != self.project.name:
                # we can't execute this query in jobs project
                raise ValueError(f"Wrong query database: {query.database}. You could try to use sql string instead")

            query = query.sql
        elif not isinstance(query, str):
            raise ValueError(f'Unable to use add this object as a query: {query}. Try to use sql string instead')
        self._queries.append(query)

    def get_history(self) -> pd.DataFrame:
        """
        Get history of job execution

        :return: dataframe with job executions
        """
        ast_query = Select(
            targets=[Star()],
            from_table=Identifier('jobs_history'),
            where=dict_to_binary_op({
                'name': self.name
            })
        )
        return self.project.api.sql_query(ast_query.to_string(), database=self.project.name)


class Jobs(CollectionBase):
    def __init__(self, project, api):
        self.project = project
        self.api = api

    def _list(self, name: str = None) -> List[Job]:

        ast_query = Select(targets=[Star()], from_table=Identifier('jobs'))

        if name is not None:
            ast_query.where = dict_to_binary_op({'name': name})

        df = self.api.sql_query(ast_query.to_string(), database=self.project.name)

        # columns to lower case
        cols_map = {i: i.lower() for i in df.columns}
        df = df.rename(columns=cols_map)

        return [
            Job(self.project, item.pop('name'), item)
            for item in df.to_dict('records')
        ]

    def list(self) -> List[Job]:
        """
        Show list of jobs in project

        :return: list of Job objects
        """

        return self._list()

    def get(self, name: str) -> Job:
        """
        Get job by name from project

        :param name: name of the job
        :return: Job object
        """

        jobs = self._list(name)
        if len(jobs) == 1:
            return jobs[0]
        elif len(jobs) == 0:
            raise AttributeError("Job doesn't exist")
        else:
            raise RuntimeError("Several jobs with the same name")

    def create(
            self,
            name: str,
            query_str: str = None,
            start_at: dt.datetime = None,
            end_at: dt.datetime = None,
            repeat_str: str = None,
            repeat_min: int = None,
        ) -> Union[Job, None]:
        """
        Create new job in project and return it.

        If it is not possible (job executed and not accessible anymore):
           return None

        Usage options:

        Option 1: to use string query
        All job tasks could be passed as string with sql queries. Job is created immediately

        >>> job = con.jobs.create('j1', query_str='retrain m1; show models', repeat_min=1):

        Option 2: to use 'with' block.
        It allows to pass sdk commands to job tasks.
        Not all sdk commands could be accepted here,
         only those which are converted in to sql in sdk and sent to /query endpoint
        Adding query sql string is accepted as well
        Job will be created after exit from 'with' block

        >>> with con.jobs.create('j1', repeat_min=1) as job:
        >>>    job.add_query(table1.insert(table2))
        >>>    job.add_query('retrain m1')  # using string

        More info about jobs: https://docs.mindsdb.com/sql/create/jobs

        :param name: name of the job
        :param query_str: str, job's query (or list of queries with ';' delimiter) which job have to execute
        :param start_at: datetime, first start of job,
        :param end_at: datetime, when job have to be stopped,
        :param repeat_str: str, optional, how to repeat job (e.g. '1 hour', '2 weeks', '3 min')
        :param repeat_min: int, optional, period to repeat the job in minutes
        :return: Job object or None
        """

        if start_at is not None:
            start_str = start_at.strftime("%Y-%m-%d %H:%M:%S")
        else:
            start_str = None

        if end_at is not None:
            end_str = end_at.strftime("%Y-%m-%d %H:%M:%S")
        else:
            end_str = None

        if repeat_min is not None:
            repeat_str = f'{repeat_min} minutes'

        def _create_callback(query):
            ast_query = CreateJob(
                name=Identifier(name),
                query_str=query,
                start_str=start_str,
                end_str=end_str,
                repeat_str=repeat_str
            )

            self.api.sql_query(ast_query.to_string(), database=self.project.name)

        if query_str is None:
            # allow to create context with job
            job = Job(self.project, name, create_callback=_create_callback)
            return job
        else:
            # create it
            _create_callback(query_str)

            # job can be executed and remove it is not repeatable
            jobs = self._list(name)
            if len(jobs) == 1:
                return jobs[0]


    def drop(self, name: str):
        """
        Drop job from project

        :param name: name of the job
        """
        ast_query = DropJob(Identifier(name))

        self.api.sql_query(ast_query.to_string(), database=self.project.name)



from dataclasses import dataclass
import dataclasses
from typing import List

from mindsdb_sql.parser.ast import Show, Identifier, BinaryOperation, Constant

from mindsdb_sdk.utils.objects_collection import CollectionBase


@dataclass(init=False)
class Handler:
    """
    :meta private:
    """
    name: str
    title: str
    version: str
    description: str
    connection_args: dict
    import_success: bool
    import_error: str

    def __init__(self, **kwargs):
        names = set([f.name for f in dataclasses.fields(self)])
        for k, v in kwargs.items():
            if k in names:
                setattr(self, k, v)


class Handlers(CollectionBase):
    """
    :meta private:
    """

    def __init__(self, api, type):
        self.api = api
        self.type = type

    def list(self) -> List[Handler]:
        """
        Returns list of handlers on server depending on type
        :return: list of handlers
        """

        ast_query = Show(
            category='HANDLERS',
            where=BinaryOperation(
                op='=',
                args=[
                    Identifier('type'),
                    Constant(self.type)
                ]
            )
        )

        df = self.api.sql_query(ast_query.to_string())
        # columns to lower case
        cols_map = {i: i.lower() for i in df.columns}
        df = df.rename(columns=cols_map)

        return [
            Handler(**item)
            for item in df.to_dict('records')
        ]

    def get(self, name: str) -> Handler:
        """
        Get handler by name

        :param name
        :return: handler object
        """
        name = name.lower()
        for item in self.list():
            if item.name == name:
                return item
        raise AttributeError(f"Handler doesn't exist: {name}")


class MLHandlers(Handlers):
    """
       **ML handlers colection**

       Examples of usage:

       Get list

       >>> con.ml_handlers.list()

       Get

       >>> openai_handler = con.ml_handlers.openai
       >>> openai_handler = con.ml_handlers.get('openai')

    """

    ...


class DataHandlers(Handlers):
    """
        **DATA handlers colection**

        Examples of usage:

        Get list

        >>> con.data_handlers.list()

        Get

        >>> pg_handler = con.data_handlers.postgres
        >>> pg_handler = con.data_handlers.get('postgres')

    """

    ...


import copy
from typing import Union
from typing import List

import pandas as pd

from mindsdb_sql.parser.ast import DropTables, CreateTable
from mindsdb_sql.parser.ast import Select, Star, Identifier, Constant, Delete, Insert, Update, Last, BinaryOperation

from mindsdb_sdk.utils.sql import dict_to_binary_op, add_condition, query_to_native_query
from mindsdb_sdk.utils.objects_collection import CollectionBase
from mindsdb_sdk.utils.context import is_saving

from .query import Query


class Table(Query):
    def __init__(self, db, name):
        # empty database
        super().__init__(db.api, '', None)
        self.name = name
        self.table_name = Identifier(parts=[db.name, name])
        self.db = db
        self._filters = {}
        self._limit = None
        self._track_column = None
        self._update_query()

    def _filters_repr(self):
        filters = ''
        if len(self._filters) > 0:
            filters = ', '.join(
                f'{k}={v}'
                for k, v in self._filters.items()
            )
            filters = ', ' + filters
        return filters

    def __repr__(self):
        limit_str = ''
        if self._limit is not None:
            limit_str = f'; limit={self._limit}'
        return f'{self.__class__.__name__}({self.table_name}{self._filters_repr()}{limit_str})'

    def filter(self, **kwargs):
        """
        Applies filters on table
        table.filter(a=1, b=2) adds where condition to table:
        'select * from table1 where a=1 and b=2'

        :param kwargs: filter
        :return: Table object
        """
        # creates new object
        query = copy.deepcopy(self)
        query._filters.update(kwargs)
        query._update_query()
        return query

    def limit(self, val: int):
        """
        Applies limit condition to table query

        :param val: limit size
        :return: Table object
        """
        query = copy.deepcopy(self)
        query._limit = val
        query._update_query()
        return query

    def track(self, column):
        """
        Apply tracking column to table. ('LAST' keyword in mindsdb)
        First call returns nothing
        The next calls return new records since previous call (where value of track_column is greater)

        Example:

        >>> query = con.databases.my_db.tables.sales.filter(type='house').track('created_at')
        >>> # first call returns no records
        >>> df = query.fetch()
        >>> # second call returns rows with created_at is greater since previous fetch
        >>> df = query.fetch()

        :param column: column to track new data from table.
        :return: Table object
        """
        query = copy.deepcopy(self)
        query._track_column = column

        query._update_query()
        return query

    def _update_query(self):
        where = dict_to_binary_op(self._filters)
        if self._track_column is not None:
            condition = BinaryOperation(op='>', args=[Identifier(self._track_column), Last()])
            where = add_condition(where, condition)

        ast_query = Select(
            targets=[Star()],
            from_table=self.table_name,
            where=where
        )
        if self._limit is not None:
            ast_query.limit = Constant(self._limit)
        self.sql = ast_query.to_string()

    def insert(self, query: Union[pd.DataFrame, Query]):
        """
        Insert data from query of dataframe
        :param query: dataframe of
        :return:
        """

        if isinstance(query, pd.DataFrame):
            # insert data
            data_split = query.to_dict('split')

            ast_query = Insert(
                table=self.table_name,
                columns=data_split['columns'],
                values=data_split['data']
            )

            sql = ast_query.to_string()

        elif isinstance(query, Query):
            # insert from select

            if query.database is not None:
                # use native query
                ast_query = Insert(
                    table=self.table_name,
                    from_select=query_to_native_query(query)
                )
                sql = ast_query.to_string()
            else:
                sql = f'INSERT INTO {self.table_name.to_string()} ({query.sql})',
        else:
            raise ValueError(f'Invalid query type: {query}')

        if is_saving():
            return Query(self, sql)

        self.api.sql_query(sql)

    def delete(self, **kwargs):
        """
        Deletes record from table using filters

        >>> table.delete(a=1, b=2)

        :param kwargs: filter
        """

        ast_query = Delete(
            table=self.table_name,
            where=dict_to_binary_op(kwargs)
        )
        sql = ast_query.to_string()

        if is_saving():
            return Query(self, sql)

        self.api.sql_query(sql)

    def update(self, values: Union[dict, Query], on: list = None, filters: dict = None):
        '''
        Update table by condition of from other table.

        If 'values' is a dict:
          it will be an update by condition
          'filters' is required
          used command: update table set a=1 where x=1

        If 'values' is a Query:
          it will be an update from select
          'on' is required
          used command: update table on a,b from (query)

        :param values: input for update, can be dict or query
        :param on: list of column to map subselect to table ['a', 'b', ...]
        :param filters: dict to filter updated rows, {'column': 'value', ...}

        '''

        if isinstance(values, Query):
            # is update from select
            if on is None:
                raise ValueError('"on" parameter is required for update from query')

            # insert from select
            if values.database is not None:
                ast_query = Update(
                    table=self.table_name,
                    keys=[Identifier(col) for col in on],
                    from_select=query_to_native_query(values)
                )
                sql = ast_query.to_string()
            else:
                map_cols = ', '.join(on)
                sql = f'UPDATE {self.table_name.to_string()} ON {map_cols} FROM ({values.sql})'

        elif isinstance(values, dict):
            # is regular update
            if filters is None:
                raise ValueError('"filters" parameter is required for update')

            update_columns = {
                k: Constant(v)
                for k, v in values.items()
            }

            ast_query = Update(
                table=self.table_name,
                update_columns=update_columns,
                where=dict_to_binary_op(filters)
            )

            sql = ast_query.to_string()
        else:
            raise NotImplementedError

        if is_saving():
            return Query(self, sql)

        self.api.sql_query(sql)


class Tables(CollectionBase):
    """
    Wortking with tables:
    Get table as Query object

    >>> table = tables.get('table1')

    Filter and limit

    >>> table = table.filter(a=1, b='2')
    >>> table = table.limit(3)

    Get content of table as dataframe. At that moment query will be sent on server and executed

    >>> df = table.fetch()

    Creating table

    From query:

    >>> table = tables.create('table2', query)

    From other table

    >>> table2 = table.create('table2', table)

    Uploading file

    >>> db = con.databases.files
    >>> db.tables.create('filename', dataframe)

  ` Droping table

    >>> db.tables.drop('table2')
    """

    def __init__(self, database, api):
        self.database = database
        self.api = api

    def _list_tables(self):
        df = self.database.query('show tables').fetch()

        # first column
        return list(df[df.columns[0]])

    def list(self) -> List[Table]:
        """
        Show list of tables in integration

        :return: list of Table objects
        """
        return [Table(self.database, name) for name in self._list_tables()]

    def get(self, name: str) -> Table:
        """
        Get table by name

        :param name: name of table
        :return: Table object
        """

        return Table(self.database, name)

    def create(self, name: str, query: Union[pd.DataFrame, Query], replace: bool = False) -> Union[Table, Query]:
        """
        Create new table and return it.

        On mindsdb server it executes command:
        `insert into a (select ...)`

        or if replace is True
        `create table a (select ...)`

        'select ...' is extracted from input Query

        :param name: name of table
        :param query: Query object
        :param replace: if true,
        :return: Table object
        """

        if isinstance(query, pd.DataFrame) and self.database.name == 'files':
            # now it is only possible for file uploading
            self.api.upload_file(name, query)

            return Table(self.database, name)

        if not isinstance(query, Query):
            raise NotImplementedError

        # # query can be in different database: wrap to NativeQuery
        # ast_query = CreateTable(
        #     name=Identifier(name),
        #     is_replace=is_replace,
        #     from_select=Select(
        #         targets=[Star()],
        #         from_table=NativeQuery(
        #             integration=Identifier(data.database),
        #             query=data.sql
        #         )
        #     )
        # )
        # self.query(ast_query.to_string()).fetch()

        # call in query database
        table = Identifier(parts=[self.database.name, name])

        if query.database is not None:
            # use native query
            ast_query = CreateTable(
                name=table,
                is_replace=replace,
                from_select=query_to_native_query(query)
            )
            sql = ast_query.to_string()
        else:
            replace_str = ''
            if replace:
                replace_str = ' or replace'

            sql = f'create{replace_str} table {table.to_string()} ({query.sql})'

        if is_saving():
            return Query(self, sql)

        self.api.sql_query(sql)

        return Table(self.database, name)

    def drop(self, name: str):
        """
        Delete table

        :param name: name of table
        """
        table = Identifier(parts=[self.database.name, name])

        ast_query = DropTables(
            tables=[table]
        )
        sql = ast_query.to_string()

        if is_saving():
            return Query(self, sql)
        self.api.sql_query(sql)




from mindsdb_sdk.server import Server

from mindsdb_sdk.connectors.rest_api import RestAPI

DEFAULT_LOCAL_API_URL = 'http://127.0.0.1:47334'
DEFAULT_CLOUD_API_URL = 'https://cloud.mindsdb.com'


def connect(
        url: str = None,
        login: str = None,
        password: str = None,
        api_key: str = None,
        is_managed: bool = False,
        headers=None) -> Server:
    """
    Create connection to mindsdb server

    :param url: url to mindsdb server
    :param login: user login, for cloud version it contents email
    :param password: user password to login (for cloud version)
    :param api_key: API key to authenticate (for cloud version)
    :param is_managed: whether or not the URL points to a managed instance
    :param headers: addtional headers to send with the connection, optional
    :return: Server object

    Examples
    --------

    >>> import mindsdb_sdk

    Connect to local server

    >>> con = mindsdb_sdk.connect()
    >>> con = mindsdb_sdk.connect('http://127.0.0.1:47334')

    Connect to cloud server

    >>> con = mindsdb_sdk.connect('https://cloud.mindsdb.com', api_key='-')

    Connect to MindsDB pro

    >>> con = mindsdb_sdk.connect('http://<YOUR_INSTANCE_IP>', login='a@b.com', password='-', is_managed=True)

    """
    if url is None:
        if login is not None:
            # default is cloud
            url = DEFAULT_CLOUD_API_URL
        else:
            # is local
            url = DEFAULT_LOCAL_API_URL

    api = RestAPI(url, login, password, api_key, is_managed, headers=headers)

    return Server(api)



from mindsdb_sql.parser.ast import BinaryOperation, Identifier, Constant, Select, Star, NativeQuery
from mindsdb_sdk.query import Query

def dict_to_binary_op(filters):
    where = None
    for name, value in filters.items():
        condition = BinaryOperation('=', args=[Identifier(name), Constant(value)])

        where = add_condition(where, condition)

    return where


def add_condition(where, condition):
    if where is None:
        return condition
    else:
        return BinaryOperation(
            'and',
            args=[where, condition]
        )


def query_to_native_query(query: Query):
    return Select(
        targets=[Star()],
        from_table= NativeQuery(
            integration=Identifier(query.database),
            query=query.sql
        )
    )


__title__ = 'mindsdb_sdk'
__package_name__ = 'mindsdb_sdk'
__version__ = '3.1.0'
__description__ = "MindsDB Python SDK, provides an SDK to use a remote mindsdb instance"
__email__ = "jorge@mindsdb.com"
__author__ = 'MindsDB Inc'
__github__ = 'https://github.com/mindsdb/mindsdb_python_sdk'
__pypi__ = 'https://pypi.org/project/mindsdb-sdk/'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright 2020- mindsdb'
